﻿////
/// Copyright(C) 2020 Micronet Co. All Rights Reserved.
//
using RTedge;
using System;
using System.Text;
using System.Diagnostics;
using System.Threading.Tasks;
using System.IO;
using System.IO.Pipes;




/// <summary>
/// --------------------------------------------------------------------------
///   RT-edge system / Container
/// 
///   EgWebSvr.exe(W)
///       OMROM PLC series data connector service
/// 
///   Copyright(c) Micronet Corp 2022, Programmed by kurakawa
/// 
/// 
///   -RT-edge service main frame module
///   -Tag pair data class
///   -WEB management class
///   
/// --------------------------------------------------------------------------
/// </summary>

namespace EgWebSvr
{


    /// <summary>
    /// タグ関連付けのアイテム
    /// </summary>
    public class TagPair
    {
        public String cmd;              // デバイスアドレス
        public EgHgTagRef EdgeTag;      // Edgeタグ
    }

    /// <summary>
    /// アウトプット用のタグ関連付けアイテム
    /// </summary>
    public class TagOutput : TagPair
    {
        public object value;      // 出力したい値

        public TagOutput()
        {
            value = null;
        }
    }


    /// <summary>
    /// サービス本体
    /// </summary>
    public partial class EdgeService
    {

        /// <summary>
        /// EdgeServiceで使用するEgWeb既定のデータ型
        /// </summary>
        public const string stTG_Base = "SERVICE.EgWebSvr,6,4,";
        public const string stTG_Status = "SERVICE.EgWebSvr.Status,3,1,";			
        public const string stTG_Error = "SERVICE.EgWebSvr.Error,1,1,";
        public const string stTG_Run = "SERVICE.EgWebSvr.Run,1,1,";
        public const string stTG_Live = "SERVICE.EgWebSvr.Live,7,4,";
        public const string stTG_PageOpenCnt = "SERVICE.EgWebSvr.PageOpenCnt,7,4,";
        public string stTG_AutoRun = "SERVICE.EgWebSvr.AutoRun,1,2,";


        public edge_API EGAPI;                                              // edge APIのインスタンス
        private System.Timers.Timer ServiceIndLiveTimer;                    // サービスインジケータ用のタイマ
        public int GB_err = 0;

        private static int numThreads = 1;


        // xmlで指定したサービスコンテナ名 EgInit後設定 デフォルト 
        public string ServiceRealName = "EgWebSvr";

        /// <summary>
        /// インジケータタグ".Run" / Service Run indicator
        /// </summary>
        bool _ServiceIndRun;
        public bool ServiceIndRun
        {
            set
            {
                _ServiceIndRun = value;
                String TagName = String.Format("SERVICE.{0}.Run", GetServiceName());
                EGAPI.EgTagWrite(TagName, value);
            }
            get { return _ServiceIndRun; }
        }

        /// <summary>
        /// インジケータタグ".Status" / Service Status indicator
        /// </summary>
        byte _ServiceIndStatus;
        byte ServiceIndStatus
        {
            set
            {
                _ServiceIndStatus = value;
                String TagName = String.Format("SERVICE.{0}.Status", GetServiceName());
                EGAPI.EgTagWrite(TagName, value);
            }
            get { return _ServiceIndStatus; }
        }

        /// <summary>
        /// インジケータタグ".Error" / Service Error indicator
        /// </summary>
        bool _ServiceIndError;
        bool ServiceIndError
        {
            set
            {
                _ServiceIndError = value;
                String TagName = String.Format("SERVICE.{0}.Error", GetServiceName());
                EGAPI.EgTagWrite(TagName, value);
            }
            get { return _ServiceIndError; }
        }

        /// <summary>
        /// インジケータタグ".Live" / ervice Live counter indicator
        /// </summary>
        UInt32 _ServiceIndLive;
        UInt32 ServiceIndLive
        {
            set
            {
                _ServiceIndLive = value;
                String TagName = String.Format("SERVICE.{0}.Live", GetServiceName());
                EGAPI.EgTagWrite(TagName, value);
            }
            get { return _ServiceIndLive; }
        }


        /// <summary>
        /// 自己のサービス名を返す
        /// </summary>
        /// <returns></returns>
        public String GetServiceName()
        {
            return ServiceRealName; //XMLで指定したサービス名 EgInit後に設定 設定がない場合はデフォルト値を返す
        }

        /// <summary>
        /// サービスプロパティのXML解釈処理
        /// </summary>
        /// <returns></returns>
        public int GetProperty()
        {
            object Value = 0;

            return 0;
        }


        /// <summary>
        /// インジケータタグの生成
        /// </summary>
        /// <returns></returns>
        public int CreateIndicater()
        {
            int ret = 0;

            // インジケータタグを生成します 自動で生成される分は指定しない
            //ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.Status", GetServiceName()) ,(ushort)EGDEFINE.egTagDataType.Byte, 1, "", "service status");
            //ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.Error", GetServiceName()), (ushort)EGDEFINE.egTagDataType.Boolean, 1, "", "service last error");
            //ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.Run", GetServiceName()), (ushort)EGDEFINE.egTagDataType.Boolean, 1, "", "service runnning");
            //ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.Live", GetServiceName()), (ushort)EGDEFINE.egTagDataType.UInt32, 4, "", "service live counter");
            ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.PageOpenCnt", GetServiceName()), (ushort)EGDEFINE.egTagDataType.UInt32, 0, "", "Web page access counter");

            return ret;
        }

        /// <summary>
        /// サービス起動時処理
        /// </summary>
        /// <returns></returns>
        public int Startup()
        {
            // RT-edgeの準備
            EGAPI = new edge_API();

            // RT-edgeユーザハンドラ登録
            edge_API.EDGECONFIG config = EGAPI.EDGE_CONFIG_DEFAULT;
            config.mInstMessFunc = MsgHandler;

            // RT-edgeフレームワーク初期化
            var ret = EGAPI.EgInit( config );
            if (ret != 0)
                edge_FW._PrintSysLog( GetServiceName(), "ERR", "Failed to initialize. Please check the environment.");

            // サービス名の設定
            //if(EGAPI.EGFW.ServiceRealName != null && EGAPI.EGFW.ServiceRealName.Length > 0) //EgInitで失敗時はnullの為チェックする
            //    ServiceRealName = EGAPI.EGFW.ServiceRealName;

            // インジケータタグの準備
            CreateIndicater();

            // XMLプロパティタグ設定の取り出し
            GetProperty();

            // MCマスタ設定のログ情報を出力
            //edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("Source IP={0} PORT={1} SNA={2} SA1={3} SA2={4}", "0.0.0.0" ,ServiceMC_PortNo , ServiceMC_SNA, ServiceMC_SA1, ServiceMC_SA2));


            // 対象PLCのリスティング,TagRefIn/TagRefOut/SystemTagの取得
            string stin = TargetPlcInputListup();
            string stout = TargetPlcOutputListup();


            string stsys = stTG_Base + "," + stTG_Status + "," + stTG_Error + "," + stTG_Run + "," + stTG_Live + "," + stTG_PageOpenCnt + "," + stTG_AutoRun;



            // ライブインジケータタグの周期更新スタート / Service Live counter timer
            ServiceIndLiveTimer = new System.Timers.Timer(1000);
            ServiceIndLiveTimer.Elapsed += OnTimedLiveIndicaterEvent;
            ServiceIndLiveTimer.Start();

            CreatePipeServerTask("EdgeWebPipe", stin, stout, stsys);

            return 0;
        }


        /// <summary>
        /// pipeサーバータスク
        /// </summary>
        /// <param name="pipeName"></param>
        public static Task CreatePipeServerTask(string pipeName, string stin, string stout, string stsys)
        {

            return Task.Run(() =>
            {
                // numThreadsで指定した数まで同じ名前のパイプを作れる
                NamedPipeServerStream pipeServer = null;
                //AnonymousPipeServerStream pipeServer = null;
                string stans = "";
                while (true)
                {
                    try
                    {
                        pipeServer = new NamedPipeServerStream(pipeName, PipeDirection.InOut, numThreads);
                        //pipeServer = new AnonymousPipeServerStream(pipeName, PipeDirection.InOut, numThreads);

                        // クライアントの接続待ち
                        pipeServer.WaitForConnection();
                        StreamString ss = new StreamString(pipeServer);

                        while (true)
                        {
                            // 受信待ち
                            var read = ss.ReadString();
                            // 受信したら応答を送信

                            
                            if (read == "TagRef_IN")
                            {
                                stans = stin;
                            }
                            if (read == "TagRef_OUT")
                            {
                                stans = stout;
                            }
                            if (read == "TagRef_SYS")
                            {
                                stans = stsys;
                            }
                            var write = ss.WriteString(stans);

                            if (read == "end") break;
                        }
                    }
                    catch (OverflowException ofex)
                    {
                        // クライアントが切断
                    }
                    finally
                    {
                        pipeServer?.Close();
                    }
                }
            });
        }

        /// <summary>
        /// Liveインジケータの更新（タイマイベント）
        /// </summary>
        /// <param name="source"></param>
        /// <param name="e"></param>
        private void OnTimedLiveIndicaterEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            try
            {
                // インジケータ更新
                ((System.Timers.Timer)source).Stop();
                ServiceIndLive++;

                // コンテナのステータスインジケータを更新、通信対象PLCのいずれかに障害があれば .Error=true にします
                int errors = 0;
                byte bst = 1;
                ServiceIndError = (errors > 0);
                //サービスのステイタスも更新する
                ServiceIndStatus = bst;
            }
            finally
            {
                // 次のタイマーイベントを許可
                ((System.Timers.Timer)source).Start();
            }
        }

        /// <summary>
        /// サービス終了時の処理
        /// </summary>
        /// <returns></returns>
        public int Cleanup()
        {

            // Liveカウンタタイマのストップ
            ServiceIndLiveTimer.Stop();

            // RT-edgeフレームワーク終了
            EGAPI.EgFinalize();

            // プロセス終了
//            Application.Exit();
            var myself = Process.GetCurrentProcess();
            myself.CloseMainWindow();
            return 0;
        }

        /// <summary>
        /// サービスメッセージハンドラ
        /// </summary>
        /// <param name="senderName"></param>
        /// <param name="messNo"></param>
        /// <param name="paramArray"></param>
        /// <returns></returns>
        public int MsgHandler(string senderName, int messNo, byte[] paramArray)
        {
            try
            {
                // 届いたEdgeメッセージに合わせて処理分岐しますよ
                switch (messNo)
                {
                    case (int)EGDEFINE.EGMSG.EM_SERVICE_RUN:            // サービス開始指令
                        ServiceIndRun = true;
                        break;
                    case (int)EGDEFINE.EGMSG.EM_SERVICE_STOP:           // サービス終了指令
                        ServiceIndRun = false;
                        Cleanup();
                        break;
                    case (int)EGDEFINE.EGMSG.EM_SERVICE_PAUSE:          // サービス一時停止指令
                        ServiceIndRun = false;
                        break;
                    case (int)EGDEFINE.EGMSG.EM_SERVICE_UPDATE:			// サービスアップデート指令
                        break;
                    default:
                        return -1;
                }
                return 0;
            }
            catch (Exception ex)
            {
                edge_FW._PrintSysLog( GetServiceName(), "ERR", "FATAL ERROR : {0}", System.Reflection.MethodBase.GetCurrentMethod().Name + " : " + ex.Message);
                return -1;
            }
        }

        // MSサンプルそのまま(streamに文字列を読み書きしてくれるクラス)
        public class StreamString
        {
            private Stream ioStream;
            private UTF8Encoding streamEncoding;
            edge_API EGAPI = new edge_API();        //APIの宣言

            public StreamString(Stream ioStream)
            {
                this.ioStream = ioStream;
                //streamEncoding = new UnicodeEncoding();
                streamEncoding = new System.Text.UTF8Encoding();
            }
            public string ReadString()
            {
                int len = 0;

                //while (true)
                //{
                //    len = ioStream.ReadByte() * 256;
                //    len += ioStream.ReadByte();
                //    if (len > 0) break; ;
                //}
                //byte[] inBuffer = new byte[len];
                //ioStream.Read(inBuffer, 0, len);

                len = ioStream.ReadByte() * 256;
                len += ioStream.ReadByte();
                byte[] inBuffer = new byte[len];
                ioStream.Read(inBuffer, 0, len);


                return streamEncoding.GetString(inBuffer);
            }

            public int WriteString(string outString)
            {
                byte[] outBuffer = streamEncoding.GetBytes(outString);
                int len = outBuffer.Length;
                if (len > UInt16.MaxValue)
                {
                    len = (int)UInt16.MaxValue;
                }
                ioStream.WriteByte((byte)(len / 256));
                ioStream.WriteByte((byte)(len & 255));
                ioStream.Write(outBuffer, 0, len);
                ioStream.Flush();

                return outBuffer.Length + 2;
            }
        }

    }
}
